
/***************************************************************************
                           ircserverconnection.cpp
                           -------------------
    copyright            : (C) 2001 by Martin Scholl
    email                : ms@deridiot.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <ircserverconnection.hpp>
#include <string>
#include <iostream>
extern "C" {
#include <sys/time.h>
}
#include <config.h>

// todo: implement usage of timeout !

static long getUNIXTimestamp()
{
  struct timeval t;
  gettimeofday(&t, NULL);

  return t.tv_sec;
}

long IRCServerConnection::DEF_TIMEOUT=600000;
string IRCServerConnection::SERVER_DESCR="ProcServ";

IRCServerConnection::IRCServerConnection(const PSInetHostAddress &addr, tpport_t port, 
					 const string &servname, const string &Password)
  : IRCConnection(addr, port),
    password(Password), hostname(servname), ircMsg(),
    timeout(DEF_TIMEOUT)
{
#ifdef VERBOSE_MEMORY
  cout << "IRCServerConnection] created" << endl;
#endif
};

IRCServerConnection::~IRCServerConnection()
{
#ifdef VERBOSE_MEMORY
  cout << "IRCServerConnection] destroyed" << endl;
#endif
}

void IRCServerConnection::onConnect()
{
#if SERVER_TYPE == SERVER_BAHAMUT
  long timestamp= getUNIXTimestamp();
#endif

#if SERVER_TYPE == SERVER_HYBRID || SERVER_TYPE == SERVER_BAHAMUT
  *tcp() << "PASS " << password << " :TS" << endl;
  *tcp() << "SERVER " << hostname << " 1 :" << SERVER_DESCR << endl;

#if SERVER_TYPE == SERVER_BAHAMUT
  *tcp() << "SVINFO 3 1 0 :" << timestamp << endl;
#endif

#endif

#ifdef DEBUG
  cout << "IRCServerConnection] PASS [***] :TS" << endl;
  cout << "IRCServerConnection] SERVER " << hostname << " 1 :HostServ" << endl;

#if SERVER_TYPE == SERVER_BAHAMUT
  cout << "IRCServerConnection] SVINFO 3 1 0 :" << timestamp << endl;
#endif

#endif

  onConnected();
};

void IRCServerConnection::onDisconnected()
{
  cout << "Disconnected from server" << endl;
}

void IRCServerConnection::onInput(char *buf, int len)
{
  cout << "IRCServerConnection] read : '";
  cout.write(buf, len);
  cout << '\'' << endl;
  ircMsg=buf;
  process(ircMsg);
};

void IRCServerConnection::process(const IrcMessage &ircMsg)
{
  const string cmd=ircMsg.getCommand();
  if(cmd.length()==4) {
    if(cmd=="PING") {
      printPong(ircMsg.getMessage());
    }    
    else if(cmd=="NICK") {
      if(ircMsg.length()==10) {
	onNickIntroduce(ircMsg.getToken(IDX_TARGET), 
			ircMsg.getToken(IDX_OPT4), 
			ircMsg.getToken(IDX_OPT5), 
			ircMsg.getToken(IDX_OPT7),
			ircMsg.getToken(IDX_OPT3),
			ircMsg.getToken(IDX_OPT2));
      }
      else if(ircMsg.length()==4) {
	onNickChange(ircMsg.getToken(IDX_HOSTMASK), ircMsg.getToken(IDX_TARGET), ircMsg.getMessage());
      }
#ifdef DEBUG
      else {
	cout << "IRCServerConnection] error: unknown message length for command NICK! Ignoring..." << endl;
      }
#endif 
    }
    else if(cmd=="MODE") {
      onMode(ircMsg);
    }
    else if(cmd=="QUIT") {
      onQuit(ircMsg);
    }
  }
  else if(cmd=="PRIVMSG") {
    onPrivmsg(ircMsg);
  }
  else if(cmd=="ERROR") {
    onError(ircMsg);
  }
  else if(cmd=="NOTICE") {
    onNotice(ircMsg);
  }
#ifdef DEBUG
  else cout << "uncatched event: " << cmd << endl;
#endif
};

//  NICK NickServ 1 1009039332 +o services ln Services.ln 0 0 :Nickname Server
void IRCServerConnection::introduceUser(const string &nick, const string &ident, 
					const string &address, const string &modes, 
					const string &realname)
{
  long timestamp=getUNIXTimestamp();
#if SERVER_TYPE == SERVER_HYBRID || SERVER_TYPE == SERVER_BAHAMUT
  *tcp() << "NICK " << nick << " 1 " << timestamp << ' ' << modes << ' ' << ident
	 << ' ' << address << ' ' << hostname; 
# ifdef DEBUG
  cout << "introduceUser(): " << "NICK " << nick << " 1 " << timestamp << ' ' << modes << ' ' << ident
       << ' ' << address << ' ' << hostname;  
# endif
  
# if SERVER_TYPE == SERVER_HYBRID 
  *tcp() << " :";
#  ifdef DEBUG
  cout << " :";
#  endif
# elif SERVER_TYPE == SERVER_BAHAMUT
  *tcp() << " 0 0 :";
#  ifdef DEBUG
  cout << " 0 0 :";
#  endif
# endif // # if SERVER_TYPE == SERVER_HYBRID

  *tcp() << realname << endl;
# ifdef DEBUG
  cout << realname << endl;
# endif // # ifdef DEBUG
#endif // #if SERVER_TYPE == SERVER_HYBRID || SERVER_TYPE == SERVER_BAHAMUT
}

void IRCServerConnection::removeUser(const string &nick)
{
  *tcp() << ':' << nick << " QUIT :" << "terminated" << endl;
#ifdef DEBUG
  cout << "removeUser(): " << ':' << nick << " QUIT :" << "terminated" << endl;
#endif
}

void IRCServerConnection::printMsg(const string &sender, const string &target, const string &msg)
{
  *tcp() << ':' << sender << ' ' << " PRIVMSG " << target << " :" << msg << endl;
#ifdef DEBUG
  cout << "printMsg(): " << ':' << sender << ' ' << " PRIVMSG " << target << " :" << msg << endl;
#endif
}

void IRCServerConnection::printNotice(const string &sender, const string &target, const string &msg)
{
  *tcp() << ':' << sender << ' ' << " NOTICE " << target << " :" << msg << endl;
#ifdef DEBUG
  cout << "printMsg(): " << ':' << sender << ' ' << " NOTICE " << target << " :" << msg << endl;
#endif
}

void IRCServerConnection::joinChannel(const string &user, const string &channel)
{
  long ts=getUNIXTimestamp();
  *tcp() << ':' << hostname << " SJOIN " << ts << ' ' << channel << " +  :" << user << endl;
#ifdef DEBUG
  cout << "joinChannel(): " << ':' << hostname << " SJOIN " << ts << ' ' << channel << " + :+" << user << endl;
#endif
}

void IRCServerConnection::quitFromServer(const char *cause)
{
  *tcp() << "SQUIT " << hostname << " :" << cause << endl;
#ifdef DEBUG
  cout << "quitFromServer(): " << "SQUIT " << hostname << " :" << cause << endl;
#endif
}

void IRCServerConnection::setDefaultServerDescription(const string &newDescr)
{
  SERVER_DESCR=newDescr;
}
