
/***************************************************************************
                           process.cpp
                           -------------------
    copyright            : (C) 2001 by Martin Scholl
    email                : ms@deridiot.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


extern "C" {
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>
}
#include <process.hpp>

Process::Process(const string &cmd, const string &args) :
  pid(0), command(cmd), arguments(args), running(false), childStdIn(0)
{
#ifdef VERBOSE_MEMORY
  cout << "Process] created: " << cmd << endl;
#endif

  init();
}

void Process::init()
{
//   setCancel(THREAD_CANCEL_IMMEDIATE);
//   setSuspend(THREAD_SUSPEND_DISABLE);
//   setSignal(SIGKILL, true);
//   setSignal(SIGTERM, true);
//   setSignal(SIGHUP, true);
}
Process::~Process()
{
  Terminate();
  if(childStdIn) fclose(childStdIn);
  cout << "Process] destroyed" << endl;
}

void Process::start()
{
  Start();
}

const string &Process::getCommand() const
{
  return command;
}

const string &Process::getArguments() const
{
  return arguments;
}

unsigned int Process::getPID() const
{
  return pid;
}

void Process::sendString(const string &s)
{
  if(!running) {
    cerr << "Process] cannot communicate with a task not running" << endl;
  }
  else if(childStdIn==0) {
    cerr << "Process] process communication setup error." << endl;
  }
  else {
#ifdef DEBUG
    cout << "Process] sending string to process " << pid << ": " << s << endl;
#endif
    if(fprintf(childStdIn, "%s\n", s.c_str())<0) perror("Process] sendString(): ");
    fflush(childStdIn);
  }
}

void Process::Initial()
{
  onStart();
}

void Process::Final()
{
  running=false;
  onTerminate();
}

void Process::Run()
{
  if(running) {
#ifdef DEBUG
    cerr << "Process] attempt to start an already running process " << endl;
#endif
    return;
  }
  else {
    running=true;
    if(!(exec())) perror("Process] error: ");
    // fclose(writeInto);
    if(errno != 29) perror("Process] starting process communication: ");
  }
}

bool Process::exec()
{
  int mypipe[2];

  if(pipe(mypipe)) {
    perror("Process] setting up pipe: ");
    return false;
  }

  pid=fork();

  if(pid==(pid_t)0) { // child
    close(mypipe[1]);
    char *args[4];
    initArgs(args, 4);
    dup2(mypipe[0], 0);
    execve("/bin/sh", args, environ);
    exit(127);
  }
  else if(pid<(pid_t)0) return false;
  else { // mother
    close(mypipe[0]);
    childStdIn=fdopen(mypipe[1], "w");
    waitpid(pid, NULL, 0);
    return true;
  }
}
 
void Process::initArgs(char *argv[], int argc)
{
  if(argc<4) argv=NULL;
  else {
    argv[0] = "sh";
    argv[1] = "-c";
    string t(command);
    t+=' ';
    t+=arguments;
    argv[2] = const_cast<char*>(t.c_str());
    argv[3] = 0;
  }
#ifdef DEBUG
  cout << "Process] arg array: argv[0]=" << argv[0] 
       << " argv[1]=" << argv[1] 
       << " argv[2]=" << argv[2]
       << endl;
#endif
  return;
}

void Process::OnSignal(int sig)
{
  switch(sig) {
  case SIGHUP: 
  case SIGKILL:
  case SIGTERM:
  case SIGINT:
  case SIGPIPE:
  case SIGSTOP:
#ifdef DEBUG
    cout << "Process] sending process " << pid << " signal: " << sig << endl;
#endif
    if(kill(pid, sig)==-1) {
      perror("Process] ");
    }
  }
}

void Process::OnHangup()
{
  OnSignal(SIGTERM);
}
