
/***************************************************************************
                           psircconnection.cpp
                           -------------------
    copyright            : (C) 2001 by Martin Scholl
    email                : ms@deridiot.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#include <psircconnection.hpp>
#include <algorithm>
extern "C" {
#include <sys/sysinfo.h>
#include <sys/types.h>
#include <linux/unistd.h>
}

const string PSIRCConnection::CMD_TASKADD=   "add";
const string PSIRCConnection::CMD_TASKHUP=   "hup";
const string PSIRCConnection::CMD_TASKDEL=   "del";
const string PSIRCConnection::CMD_TASKKILL=  "kill";
const string PSIRCConnection::CMD_TASKCOMM=  "comm";
const string PSIRCConnection::CMD_SYSINFO=   "sysinfo";
const string PSIRCConnection::CMD_AUTH=      "auth";
const string PSIRCConnection::CMD_DEAUTH=    "deauth";
const string PSIRCConnection::CMD_REHASH=    "rehash";
const string PSIRCConnection::CMD_LISTVPROGS="listvprogs";
const string PSIRCConnection::CMD_HELP=      "help";
string PSIRCConnection::MANAGER_NICK=        "ProcServ";
string PSIRCConnection::MANAGER_IDENT=       "proc";
string PSIRCConnection::MANAGER_ADDRESS=     "serv";
string PSIRCConnection::MANAGER_MODES=       "+o";
string PSIRCConnection::MANAGER_REALNAME=    "ProcServ .06";

/**
   todo: implement processAuth()
*/

PSIRCConnection::PSIRCConnection( const PSInetHostAddress &addr, 
				  tpport_t port,
				  const string &servname, 
				  const string &password,
				  VProgs *Vprogs,
				  Logins *AuthFile )
  : IRCServerConnection(addr, port, servname, password),
    ircClients(new ClientList()), tasks(new ClientTaskSet()),
    vprogs(Vprogs), authData(AuthFile),
    defaultUserName(MANAGER_NICK)
{
#ifdef VERBOSE_MEMORY
  cout << "PSIRCConnection] created" << endl;
  if(!vprogs) cout << "PSIRCConnection] warning: no vprogs given" << endl;
  if(!authData) cout << "PSIRCConnection] warning: no logins given" << endl;
#endif
}

PSIRCConnection::~PSIRCConnection()
{
  delete ircClients;
  delete tasks;

#ifdef VERBOSE_MEMORY
  cout << "PSIRCConnection] destroyed" << endl;
#endif
}

void PSIRCConnection::setDefaultManagerNick(const string &nick)
{
  MANAGER_NICK=nick;
}

void PSIRCConnection::setDefaultManagerIdent(const string &ident)
{
  MANAGER_IDENT=ident;
}

void PSIRCConnection::setDefaultManagerAddress(const string &address)
{
  MANAGER_ADDRESS=address;
}

void PSIRCConnection::setDefaultManagerModes(const string &modes)
{
  MANAGER_MODES=modes;
}

void PSIRCConnection::setDefaultManagerRealname(const string &realname)
{
  MANAGER_REALNAME=realname;
}

void PSIRCConnection::setDefaultManagerUserMode(const string &modeString)
{
  MANAGER_MODES=modeString;
}

void PSIRCConnection::onPrivmsg(const IrcMessage &msg) 
{
  const string &target=msg.getToken(IDX_TARGET);
  if( containsIgnoreCase( target.begin(), target.end(), defaultUserName.begin(), defaultUserName.end() ) ) {
    processManagerMsg(msg);
  }
  else {
    processProcessMsg(msg);
  }
}

void PSIRCConnection::onNotice(const IrcMessage &msg) 
{
  onPrivmsg(msg); // :-)
}

void PSIRCConnection::onConnected()
{
  introduceUser(defaultUserName, MANAGER_IDENT, MANAGER_ADDRESS, 
		MANAGER_MODES, MANAGER_REALNAME);
}

void PSIRCConnection::onError(const IrcMessage &msg)
{
  stopPolling();
}

void PSIRCConnection::onQuit(const IrcMessage &msg)
{
  const string &target=msg.getToken(IDX_HOSTMASK);
  Client *c=ircClients->getClient(target);
  if(c) {
#ifdef DEBUG
    cout << "PSIRCConnection] removing user entry for " << target << endl;
#endif
    delete c;
  }
#ifdef DEBUG
  else cout << "PSIRCConnection] client object not found for user: " << target << endl;
#endif
}

void PSIRCConnection::onNickChange(const IrcToken &oldNick,
				   const IrcToken &newNick,
				   const IrcToken &timestamp)
{
  Client *c=ircClients->getClient( oldNick );
  if(c) c->setNickname(newNick);
#ifdef DEBUG
  else cout << "PSIRCConnection] could not find user " << oldNick << endl;
#endif
}

void PSIRCConnection::setNewAuthData( Logins *l )
{
  authData = l;
}

void PSIRCConnection::setNewVProgs( VProgs *p )
{
  vprogs = p ;
}

// !!!!!!!!!!!!!!!!!!!!!!
// !!!!!!!improve !!!!!!!
// !!!!!!!!!!!!!!!!!!!!!!
void PSIRCConnection::processManagerMsg(const IrcMessage &ircMsg)
{
  const string & target = ircMsg.getToken(IDX_HOSTMASK);
  string tokens[3];
  Client *targetClientObj=ircClients->getClient( target );
  bool authedUser = targetClientObj != NULL;

  int tokCount=fillTokenArray(ircMsg.getMessage(), &tokens[0], 3);
  const string & cmd=tokens[0];

  if( tokCount==0 ) printManagerHelp(target);
  else if( containsIgnoreCase( cmd.begin(), cmd.end(), CMD_AUTH.begin(), CMD_AUTH.end() ) ) {
    if( tokCount<2 ) printManagerHelp( target );
    else if( authedUser ) printAlreadyLogedin( target );
    else processAuth( target, tokens[1], tokens[2] );
  } 
  // everything following from here requires an account at procserv
  else if( !authedUser ) { 
    printLoginFirst( target );
  }
  else if(containsIgnoreCase( cmd.begin(), cmd.end(), CMD_DEAUTH.begin(), CMD_DEAUTH.end() ) ) {
    delete targetClientObj;
    printDeAuthDone( target );
  }
  else if( containsIgnoreCase( cmd.begin(), cmd.end(), CMD_TASKADD.begin(), CMD_TASKADD.end() ) ) {
    if(tokCount<1) {
      printManagerHelp(target);
      return;
    }
    processTaskAdd( target, tokens[1], tokens[2], tokCount == 3 );
  }
  else if( containsIgnoreCase( cmd.begin(), cmd.end(), CMD_SYSINFO.begin(), CMD_SYSINFO.end() ) ) {
    printSysInfo( target );
  }
  else if( containsIgnoreCase( cmd.begin(), cmd.end(), CMD_REHASH.begin(), CMD_REHASH.end() ) ) {
    processRehash( target );
  }
  else if( containsIgnoreCase( cmd.begin(), cmd.end(), CMD_LISTVPROGS.begin(), CMD_LISTVPROGS.end() ) ) {
    processListVProgs( target );
  }
  else {
    if( ! containsIgnoreCase( cmd.begin(), cmd.end(), CMD_HELP.begin(), CMD_HELP.end() ) ) printNoSuchCommand( target );
    printManagerHelp( target );
  }
}

void PSIRCConnection::processProcessMsg(const IrcMessage &ircMsg)
{
  const string & sender=ircMsg.getToken(IDX_HOSTMASK);  
  PSTask * task=const_cast< PSTask* >( (*tasks)[ircMsg.getToken(IDX_TARGET)] );
  if(task==NULL) { // debugging
#ifdef DEBUG
    cout << "PSIRCConnection] errrrr no such task" << endl;
#endif
    return;
  }
  
  string tokens[2];
  int tokCount=fillTokenArray(ircMsg.getMessage(), &tokens[0], 2); 
 
  const string & command=tokens[0];
  if(command.length()>2 && command.length() <5) {
    if( containsIgnoreCase( command.begin(), command.end(), CMD_TASKCOMM.begin(), CMD_TASKCOMM.end() ) ) {
      if( tokCount>1 ) processTaskComm( task, tokens[1] );
      else printTaskHelp( sender );
    }
    else if( containsIgnoreCase( command.begin(), command.end(), CMD_TASKDEL.begin(), CMD_TASKDEL.end() ) ) {
      processTaskDel( task );
    }
    else if( containsIgnoreCase( command.begin(), command.end(), CMD_TASKKILL.begin(), CMD_TASKKILL.end() ) ) {
      processTaskKill( task );
    }
    else if( containsIgnoreCase( command.begin(), command.end(), CMD_TASKHUP.begin(), CMD_TASKHUP.end() ) ) {
      processTaskHup( task );
    }
    else {
      if( ! containsIgnoreCase( command.begin(), command.end(), CMD_HELP.begin(), CMD_HELP.end() ) ) printNoSuchCommand( sender );
      printTaskHelp( sender );
    }
  }
  else {
    if( ! containsIgnoreCase( command.begin(), command.end(), CMD_HELP.begin(), CMD_HELP.end() ) ) printNoSuchCommand( sender );
    printTaskHelp( sender );
  }
}

void PSIRCConnection::processTaskAdd(const string &target, const string &processType, const string &ircArgs, bool ircArgsSet)
{
  cout << "processTaskAdd(" << target << ", " << processType << ", " << ircArgs << ", " << ircArgsSet << ')' << endl;
  const ProcessAlias *ps=(*vprogs)[processType];
  if( ps ) {    
    PSTask *t = NULL;
    if( ircArgsSet && ps->overwriteOptions() ) {
#ifdef DEBUG
      cout << "PSIRCConnection] using additiotional arguments" << endl;
#endif
      string parsedArgs = parseProcessArgs( ps->getOptions(), ircArgs );
      t = new PSTask( const_cast<PSIRCConnection*>( this ), tasks, ps, parsedArgs, true );
    }
    else      
      t = new PSTask( const_cast<PSIRCConnection*>( this ), tasks, ps, ps->getOptions(), true );
  }
  else printNoSuchVProg(target, processType);
}

void PSIRCConnection::processTaskDel( PSTask *t )
{
  if( t ) t->terminate(); // initiate destruction process - pstask is selfdeleting
#ifdef DEBUG
  else cerr << "PSIRCConnection] errrrrr Task *t==NULL" << endl;
#endif
}

void PSIRCConnection::processTaskKill( PSTask *t )
{
  if( t ) t->kill();
#ifdef DEBUG
  else cerr << "PSIRCConnection] errrrrr Task *t==NULL" << endl;
#endif
}

void PSIRCConnection::processTaskHup( PSTask *t )
{
  if( t ) t->reload();
#ifdef DEBUG
  else cerr << "PSIRCConnection] errrrrr Task *t == NULL" << endl;
#endif
}

void PSIRCConnection::processTaskComm(PSTask *t, const string &msg)
{
  if( msg.length()==0 ) {
#ifdef DEBUG
    cout << "PSIRCConnection] trying to send empty string to process. ignoring" << endl;
#endif
    return;
  }
  if(t) t->sendString(msg);
#ifdef DEBUG
  else cerr << "PSIRCConnection] errrrrr Task *t==NULL" << endl;
#endif
}

void PSIRCConnection::processAuth(const string &sender, const string &username, const string &password)
{
  if((*authData)[username]==password) {
    new Client(ircClients, sender, username, password, sender);
    printAuthDone( sender );
  }
  else printAuthFailed( sender );

}

void PSIRCConnection::processRehash(const string &sender)
{
  kill( getpid(), SIGHUP ); // emit SIGHUP to main process
  printRehashed( sender );
}

void PSIRCConnection::processListVProgs(const string &sender)
{
  for( VProgs::const_iterator it=vprogs->begin(); it != vprogs->end(); ++it ) {
    printVProg( it->first, *(it->second), sender );
  }
}

bool nocase_compare(const char &c1, const char &c2)
{
  return toupper(c1) == toupper(c2);
}

bool PSIRCConnection::containsIgnoreCase(string::const_iterator hbegin,
					 string::const_iterator hend,
					 string::const_iterator nbegin, 
					 string::const_iterator nend)
{
  string::const_iterator it=search(hbegin, hend, nbegin, nend, nocase_compare);
  return it == hbegin;
}

void PSIRCConnection::printNoSuchCommand(const string &sender)
{
  const static string ERROR_MSG = "No such command";
  printNotice( defaultUserName, sender, ERROR_MSG );
}

void PSIRCConnection::printLoginFirst( const string &target )
{
  const static string HELP_MSG1 = "You have to login first.";
  const static string HELP_MSG2 = "Available commands to you are:";
  printNotice( defaultUserName, target, HELP_MSG1 );
  printNotice( defaultUserName, target, HELP_MSG2 );
  printLoginHelp( target );
}

void PSIRCConnection::printTaskHelp(const string &target)
{
  static const string THELP  = "Available commands for ProcServ tasks:";
  static const string THELP1 = CMD_TASKCOMM + " [string(s)] - communicate with task";
  static const string THELP2 = CMD_TASKDEL + "              - stop task  ('CTRL+C')";
  static const string THELP3 = CMD_TASKHUP + "              - send signal SIGHUP to task ('reload')";
  static const string THELP4 = CMD_TASKKILL + "             - stop task immediately (sig 9)";

  printNotice( defaultUserName, target, THELP  );
  printNotice( defaultUserName, target, THELP1 );
  printNotice( defaultUserName, target, THELP2 );
  printNotice( defaultUserName, target, THELP3 );
  printNotice( defaultUserName, target, THELP4 );
}

void PSIRCConnection::printManagerHelp(const string &target)
{
  static const string MHELP  = "Available ProcServ manager commands :";
  static const string MHELP2 = CMD_DEAUTH+ "                     - deauthenticate";
  static const string MHELP3 = CMD_TASKADD + " <type> <args>          - adds a process of type <type> and arguments <args>"; 
  static const string MHELP4 = CMD_SYSINFO + "                    - view some basic runtime information about the host procserv is running on."; 
  static const string MHELP5 = CMD_REHASH + "                     - reload configuration (hopefully ;)";
  static const string MHELP6 = CMD_LISTVPROGS + "                 - list _all_ process aliases";

  printNotice( defaultUserName, target, MHELP  );
  printLoginHelp( target );
  printNotice( defaultUserName, target, MHELP2 );
  printNotice( defaultUserName, target, MHELP3 );
  printNotice( defaultUserName, target, MHELP4 );
  printNotice( defaultUserName, target, MHELP5 );
  printNotice( defaultUserName, target, MHELP6 );
}

int PSIRCConnection::fillTokenArray(const string &msg, string *target, int len, char delim)
{
  string::size_type begin=0, end=0;
  string *array=target;

  int i=0;
  for(; i<len; ++i) {
    begin=msg.find_first_not_of(delim, end);
    if(begin==string::npos) break;
    
    end=msg.find(delim, begin);
    if(end==string::npos) {
      *array=msg.substr(begin);
      ++i;
      break;
    }
    *array=msg.substr(begin, end-begin);
    ++array;
  }
  // clear every remaining token in array
  for( int j = i; (++j) < len; ) (++target)->erase();

  return i;
}

void PSIRCConnection::printSysInfo(const string &target)
{
  cout << "PSIRCConnection] printing sysinfo to " << target << endl;
  struct sysinfo info;
  if( sysinfo(&info) ) {
    *tcp() << ':' << defaultUserName << " NOTICE " << target << " :"
	   << "sysinfo not available" << endl;
    return;
  }

  *tcp() << ':' << defaultUserName << " NOTICE " << target << " :" 
	 << "Load: " 
	 << (info.loads[0]/100000) << '.' << (info.loads[0]%100000) << "(1 min) "
	 << (info.loads[1]/100000) << '.' << (info.loads[1]%100000) << "(5 min) " 
	 << (info.loads[2]/100000) << '.' << (info.loads[2]%100000) << "(15 min)" 
	 << endl;

  *tcp() << ':' << defaultUserName << " NOTICE " << target << " :" 
	 << "Ram: " 
	 << info.totalram/1024 << " kb (total) " 
	 << info.freeram/1024 << " kb (free) " 
	 << info.sharedram/1024 << " kb (shared) " 
	 << info.bufferram/1024 << " kb (buffer) "
	 << (info.totalram-info.freeram-info.sharedram-info.bufferram)/1024 << "(used)"
	 << endl;

  *tcp() << ':' << defaultUserName << " NOTICE " << target << " :"
	 << "Swap: " 
	 << info.totalswap/1024 << " kb (total) " 
	 << info.freeswap/1024 << " kb (free) "
	 << (info.totalswap-info.freeswap)/1024 << " kb (used)" << endl;
}

void PSIRCConnection::printDeAuthFailed( const string &target)
{
  const static string ERROR_MSG = "deauth failure";

  printNotice( defaultUserName, target, ERROR_MSG );
}

void PSIRCConnection::printDeAuthDone( const string &target ) 
{
  const static string MSG = "deauth done";

  printNotice( defaultUserName, target, MSG );
}

void PSIRCConnection::printAuthDone( const string &target )
{
  const static string MSG = "login done";

  printNotice( defaultUserName, target, MSG );
}

void PSIRCConnection::printAuthFailed( const string &target )
{
  const static string ERROR_MSG = "auth failure";

  printNotice( defaultUserName, target, ERROR_MSG );
}

void PSIRCConnection::printLoginHelp( const string &target )
{
  static const string HELP_MSG1 = CMD_AUTH + " <username> <password> - authenticate yourself at procserv";
  printNotice( defaultUserName, target, HELP_MSG1);
}

void PSIRCConnection::printAlreadyLogedin( const string &target )
{
  const static string ERROR_MSG="You have already loged in. Logout first.";
  printNotice( defaultUserName, target, ERROR_MSG );
}

void PSIRCConnection::printRehashed( const string &target )
{
  const static string NOTICE_MSG="Successfully rehashed...";
  printNotice( defaultUserName, target, NOTICE_MSG );
}

void PSIRCConnection::printVProg(const string &vName, const ProcessAlias &alias, 
				 const string &ircTarget)
{
  *tcp() << ':' << defaultUserName << " NOTICE " << ircTarget << " :" << vName << " -> { " << alias << " }" << endl;
}

void PSIRCConnection::printNoSuchVProg(const string &ircTarget, const string &vprogname)
{
  *tcp() << ':' << defaultUserName << " NOTICE " << ircTarget << " :No such vprog: "  << vprogname << endl;
}

string PSIRCConnection::parseProcessArgs(const string &origArgs, const string &optArgs)
{
  string dest;
  if( optArgs[0] == '+' || optArgs[0] == '/' ) {
    dest=origArgs;
    string::const_iterator beg = optArgs.begin();
    dest.append( ++beg, optArgs.end() );
  }
  else dest = optArgs;

  return dest;
}
