#include <vprogs.hpp>
#include <iostream>
#include <algorithm>

const string VProgs::XMLENTITY_NAME                = "Processes";
const string VProgs::XMLENTITY_OVERWRITEOPTIONS    = "OverwriteArguments";
const string VProgs::XMLENTITY_DEFAULTIRCIDENT     = "DefaultIRCIdent";
const string VProgs::XMLENTITY_DEFAULTIRCADDRESS   = "DefaultIRCAddress";
const string VProgs::XMLENTITY_DEFAULTIRCREALNAME  = "DefaultIRCRealname";
const string VProgs::XMLENTITY_DEFAULTIRCMODES     = "DefaultIRCModes";
const string VProgs::XMLENTITY_DEFAULTIRCCHANNEL   = "DefaultIRCChannel";
const string VProgs::XMLENTITY_PROCESS_NAME        = "Process";
const string VProgs::XMLENTITY_PROCESS_ID          = "ID";
const string VProgs::XMLENTITY_PROCESS_COMMAND     = "Command";
const string VProgs::XMLENTITY_PROCESS_OPTIONS     = "Options";
const string VProgs::XMLENTITY_PROCESS_IRCNAME     = "IRCName";
const string VProgs::XMLENTITY_PROCESS_IRCIDENT    = "IRCIdent";
const string VProgs::XMLENTITY_PROCESS_IRCREALNAME = "IRCRealname";
const string VProgs::XMLENTITY_PROCESS_IRCADDRESS  = "IRCAddress";
const string VProgs::XMLENTITY_PROCESS_IRCMODES    = "IRCModes";
const string VProgs::XMLENTITY_PROCESS_IRCCHANNEL  = "IRCChannel";
const string VProgs::XMLENTITY_PROCESS_OVERWRITEOPTIONS = "OverwriteArguments";

VProgs::VProgs()
  : map<const string, const ProcessAlias *> ()
{
#ifdef VERBOSE_MEMORY
  cout << "VProgs] created" << endl;
#endif
}

VProgs::VProgs( const XMLEntity *c ) 
  : map<const string, const ProcessAlias*> ()
{
#ifdef VERBOSE_MEMORY
  cout << "VProgs] created" << endl;
#endif
  process( c );
}

VProgs::~VProgs()
{
  cleanUp();
#ifdef VERBOSE_MEMORY
  cout << "VProgs] destroyed" << endl;
#endif
}

bool VProgs::process( const XMLEntity *entity )
{
  bool ret = false;
  if( entity->getType() == XMLENTITY_NAME ) {
    processXMLEntity( *(const_cast<XMLEntity*>( entity )) );
    if( !entity->getSubEntities().empty() ) {
      processSubXMLEntities(entity->getSubEntities().begin(),
			    entity->getSubEntities().end());
      ret=true;
    }
    else {
      cout << "Warning: No vprogs defined!" << endl;
      ret=false;
    }
  }

  return ret;
}

void VProgs::processXMLEntity ( XMLEntity &entity )
{
  const string &defIrcIdent    = entity[XMLENTITY_DEFAULTIRCIDENT];
  if( defIrcIdent.length() != 0 ) ProcessAlias::setDefaultIdent( defIrcIdent );
  
  const string &defIrcAddress  = entity[XMLENTITY_DEFAULTIRCADDRESS];
  if( defIrcAddress.length() != 0 ) ProcessAlias::setDefaultAddress( defIrcAddress );

  const string &defIrcRealname = entity[XMLENTITY_DEFAULTIRCREALNAME];
  if( defIrcRealname.length() != 0 ) ProcessAlias::setDefaultRealname( defIrcRealname );

  const string &defIrcModes    = entity[XMLENTITY_DEFAULTIRCMODES];
  if( defIrcModes.length() != 0 ) ProcessAlias::setDefaultModes( defIrcModes );

  string &defIrcChannel  = entity[XMLENTITY_DEFAULTIRCCHANNEL];
  if( defIrcChannel.length() != 0) ProcessAlias::setDefaultChannel( defIrcChannel );

  const string &overwrite      = entity[XMLENTITY_OVERWRITEOPTIONS];
  if( overwrite.length() > 0 ) { 
    bool overwriteOptions = atoi( overwrite.c_str() ) > 0;
    ProcessAlias::setOverwriteOptions(overwriteOptions );
  }
}

void VProgs::processSubXMLEntities(XMLEntity::SubEntitiesList::const_iterator begin, 
				   XMLEntity::SubEntitiesList::const_iterator end)

{
  for(XMLEntity::SubEntitiesList::const_iterator it = begin; it != end; ++it ) {
    if((*it)->getType()==XMLENTITY_PROCESS_NAME) {
      if(checkXMLProcessEntity(*(const_cast<XMLEntity*>(*it)))) {
	insertXMLProcessEntity(*(*it));
      }
#ifdef DEBUG
      else {
	cout << "VProgs] Skipping entity : " << (*it)->getType() << endl;
      }
#endif
    }  
  }
}

bool VProgs::checkXMLProcessEntity( XMLEntity &entity )
{
  return entity[XMLENTITY_PROCESS_ID].length() != 0 && 
    entity[XMLENTITY_PROCESS_COMMAND].length() != 0 ;
}

void VProgs::insertXMLProcessEntity( XMLEntity &entity )
{
  const string &name = entity[XMLENTITY_PROCESS_ID];

  const string &command = entity[XMLENTITY_PROCESS_COMMAND];

  const string &options = entity[XMLENTITY_PROCESS_OPTIONS];

  string ircname = entity[XMLENTITY_PROCESS_IRCNAME];
  if( ircname.length() == 0 )     ircname     = name;

  string ircIdent = entity[XMLENTITY_PROCESS_IRCIDENT];
  if( ircIdent.length() == 0 )    ircIdent    = ProcessAlias::getDefaultIdent();

  string ircAddress = entity[XMLENTITY_PROCESS_IRCADDRESS];
  if( ircAddress.length() == 0 )  ircAddress  = ProcessAlias::getDefaultAddress();

  string ircRealname = entity[XMLENTITY_PROCESS_IRCREALNAME];
  if( ircRealname.length() == 0 ) ircRealname = ProcessAlias::getDefaultRealname();

  string ircModes = entity[XMLENTITY_PROCESS_IRCMODES];
  if( ircModes.length() == 0 )    ircModes    = ProcessAlias::getDefaultModes();

  string ircChannel = entity[XMLENTITY_PROCESS_IRCCHANNEL];
  if( ircChannel.length() == 0 )  ircChannel  = ProcessAlias::getDefaultChannel();

  const string &ovrwrite = entity[XMLENTITY_PROCESS_OVERWRITEOPTIONS];
  bool overwrite=ProcessAlias::getDefaultOverwriteOptions();
  if(ovrwrite.length() > 0 ) overwrite = atoi( ovrwrite.c_str() ) != 0;
  
  ProcessAlias *ps=new ProcessAlias( command, options, 
				     ircname, ircIdent, ircAddress, 
				     ircRealname, ircModes, ircChannel,
				     overwrite );

  insert( make_pair( name, ps ) );

#ifdef DEBUG
  cout << "VProgs] inserted: " << *ps << endl;
#endif
}

void deletePtr( pair<const string, const ProcessAlias *> &p )
{
  delete p.second;
}


void VProgs::cleanUp()
{

}

